from __future__ import division, print_function, absolute_import

import matplotlib

matplotlib.use('Agg')
from tensorflow.python.util import deprecation
import matplotlib.pyplot as plt
import tensorflow as tf
import os
import cv2 as cv2
import glob as glob
from tools_resnet import encoder, decoder, latent_space, vae_loss
from numpy import save
from matplotlib.offsetbox import OffsetImage

deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.contrib._warning = None
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = "6"

image_dim = 128
# Training parameters
learning_rate = 1e-5
batch_size = 32
latent_dim = 64
# Saving and loading parameters
location_to_save = '/src/models'
path_to_decoder = '/src/datasets/bestMse2/bestmodel.ckpt'
path_to_dataset = '/src/datasets/FINAL2'

# Charge data
entries = os.listdir(path_to_dataset)
neutral_faces = []
neutral_paths = []
for directory in entries:
    filenames = [img for img in glob.glob(path_to_dataset + '/' + directory + '/*.png')]
    count = 0
    neutral_found = False
    neutral = ''
    for file in filenames:
        if os.path.basename(file)[0] == 'N':
            neutral = file
            neutral_found = True
    if neutral_found :
        neutral_faces.append(cv2.imread(neutral) / 255.)

print(len(neutral_faces))
##########################################   MODEL   ##########################################
# Placeholders
X_true = tf.placeholder("float", [None, image_dim, image_dim, 3])
Z_sampled = tf.placeholder("float", [None, latent_dim])
beta = tf.placeholder(tf.float32, [], name='beta')
d_rate = tf.placeholder_with_default(1.0, shape=(), name='dropout')
# Encode
enco = encoder(X_true, d_rate)
# Generate distribution
z_mean, z_std = latent_space(enco, latent_dim)
eps = tf.random_normal(tf.shape(z_std), dtype=tf.float32, mean=0., stddev=1.0, name='epsilon')
# Sampler: Normal (gaussian) random distribution
# Reparametrization trick
Z_sampled = z_mean + tf.exp(z_std / 2.) * eps
# Decode
deco = decoder(Z_sampled, d_rate)
# Loss
loss, reconstruction_loss, kdloss = vae_loss(deco, X_true, beta, z_mean, z_std)
# Optimizer
train_op = tf.train.AdamOptimizer(learning_rate).minimize(loss)
# Initialize variables
init = tf.global_variables_initializer()
##########################################   FINISH MODEL   #######################################

def getImage(path):
    return OffsetImage(plt.imread(path), zoom=0.25)

# Start Training
# Start a new TF session
with tf.Session() as sess:
    # Run the initializer
    sess.run(init)
    reuse_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES,
                                   scope="vae[0123456789]")
    reuse_vars_dict = dict([(var.op.name, var) for var in reuse_vars])
    restore_saver = tf.train.Saver(reuse_vars_dict)
    restore_saver.restore(sess, path_to_decoder)

    total_batchs = int(len(neutral_faces) / batch_size)
    print("Total Batch Size : ", total_batchs)
    for i in range(total_batchs):
        neutral_batch = neutral_faces[i * batch_size:((i + 1) * batch_size)]
        entries_tosave = entries[i * batch_size:((i + 1) * batch_size)]
        z = sess.run(Z_sampled,feed_dict={X_true:neutral_batch})
        i = 0
        for dir in entries_tosave:
            save(path_to_dataset + '/' + dir + '/data_incr2.npy', z[i])
            i = i + 1

    print('---> FINISHED_SAVING')
