from __future__ import division, print_function, absolute_import
from tensorflow.examples.tutorials.mnist import input_data
from tensorflow.python.util import deprecation
import matplotlib.pyplot as plt
import tensorflow as tf
import numpy as np
import os

ksize = 3

# Define VAE Loss
def vae_loss(x_reconstructed, x_true, beta, z_mean, z_std,z,reg_v):
    # Reconstruction loss
    x_reconstructed = tf.layers.flatten(x_reconstructed)
    x_true = tf.layers.flatten(x_true)
    l2_loss = tf.reduce_sum(tf.pow(x_true - x_reconstructed, 2), 1)
    reg_loss = tf.reduce_sum(tf.pow(z - reg_v, 1), 1)
    return tf.reduce_mean(l2_loss), tf.reduce_mean(l2_loss), tf.reduce_mean(tf.abs(reg_loss))


def file_to_print(dirName):
    try:
        # Create target Directory
        os.mkdir(dirName)
        os.mkdir(dirName + '/recons')
        os.mkdir(dirName + '/plots')
        os.mkdir(dirName + '/save')
        print("Directory ", dirName, " Created ")
        f = open(dirName + "/log.txt", "w+")
        # info = 'con batch norm en to do, si restnet , max cada 3 , 32-512(3), relu , 1024 en fc '
        # f.write(info + '\n')
    except FileExistsError:
        print("Directory ", dirName, " already exists")
    return f


def convo_layer_max(input, nfilters, ksize, strides):
    layer_norm = tf.layers.conv2d(input, nfilters, ksize, padding="SAME")
    max_layer = tf.nn.max_pool(layer_norm, ksize=[1, ksize, ksize, 1], strides=[1, strides[0], strides[1], 1],
                               padding='SAME')
    return max_layer


def fc_layer(input, num_outputs, rate):
    layer_norm = tf.layers.dense(input, num_outputs)
    drop = tf.nn.dropout(layer_norm, rate)
    activation_layer = tf.nn.elu(drop)
    return activation_layer


def fc_layer_nodrop(input, num_outputs):
    layer_norm = tf.layers.dense(input, num_outputs)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def fc_layer_nodrop_deco(input, num_outputs,name,train):
    layer_norm = tf.layers.dense(input, num_outputs,name=name,trainable=train)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def fc_layer_deco(input, num_outputs, rate,name,train):
    layer_norm = tf.layers.dense(input, num_outputs,name=name,trainable=train)
    drop = tf.nn.dropout(layer_norm, rate)
    activation_layer = tf.nn.elu(drop)
    return activation_layer


def convo_layer(input, nfilters, ksize, strides):
    layer_norm = tf.layers.conv2d(input, nfilters, ksize, strides=strides, padding="SAME")
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def unconvo_layer(input, nfilters, ksize, strides,name,train):
    layer_norm = tf.layers.conv2d_transpose(input, nfilters, ksize, strides=strides, padding="SAME",name=name,trainable=train)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def encoder(input, d_rate,num_images):
    input = tf.reshape(input, shape=[-1, 128, 128, 3])

    conv0 = convo_layer(input, 64, ksize, (1, 1))
    conv1 = convo_layer(conv0, 64, ksize, (1, 1))
    conv2 = convo_layer_max(conv1, 128, 2, (2, 2))
    conv3 = convo_layer(conv2, 128, ksize, (1, 1))
    conv4 = convo_layer(conv3, 128, ksize, (1, 1))
    conv5 = convo_layer_max(conv4, 256, 2, (2, 2))
    conv6 = convo_layer(conv5, 256, ksize, (1, 1))
    conv7 = convo_layer(conv6, 256, ksize, (1, 1))
    conv8 = convo_layer(conv7, 256, ksize, (1, 1))
    conv9 = convo_layer_max(conv8, 512, 2, (2, 2))
    conv10 = convo_layer(conv9, 512, ksize, (1, 1))
    conv11 = convo_layer(conv10, 512, ksize, (1, 1))
    conv12 = convo_layer(conv11, 512, ksize, (1, 1))
    conv13 = convo_layer_max(conv12, 512, 2, (2, 2))
    conv14 = convo_layer(conv13, 512, ksize, (1, 1))
    conv15 = convo_layer(conv14, 512, ksize, (1, 1))
    conv16 = convo_layer(conv15, 512, ksize, (1, 1))
    conv17 = convo_layer_max(conv16, 512, 2, (2, 2))

    avg = tf.compat.v1.layers.flatten(conv17)
    avg = tf.reshape(avg, shape=[-1, num_images, 8192])
    unified_tensor = tf.math.reduce_mean(avg, axis=1)

    fc0 = fc_layer(unified_tensor, 4096, d_rate)
    fc1 = fc_layer(fc0, 4096, d_rate)
    fc2 = fc_layer_nodrop(fc1, 1024)

    return fc2


def latent_space(input, latent_dim):
    mean_ = tf.layers.dense(input, units=latent_dim)  # , activation=tf.nn.sigmoid
    std_dev = tf.nn.softplus(tf.layers.dense(input, units=latent_dim)) + 1e-6
    return mean_, std_dev


def decoder(input, d_rate):
    train = True
    fc0 = fc_layer_nodrop_deco(input, 1024,'vae24',train)
    fc1 = fc_layer_deco(fc0, 4096, d_rate,'vae25',train)
    fc2 = fc_layer_deco(fc1, 4096, d_rate,'vae26',train)

    fc3 = fc_layer_nodrop_deco(fc2, 4 * 4 * 512,'vae27',train)
    rsl = tf.reshape(fc3, shape=[-1, 4, 4, 512])

    unconv0 = unconvo_layer(rsl, 512, 2, (2, 2),'vae28',train)
    unconv1 = unconvo_layer(unconv0, 512, ksize, (1, 1),'vae29',train)
    unconv2 = unconvo_layer(unconv1, 512, ksize, (1, 1),'vae30',train)
    unconv3 = unconvo_layer(unconv2, 512, ksize, (1, 1),'vae31',train)
    unconv4 = unconvo_layer(unconv3, 512, 2, (2, 2),'vae32',train)
    unconv5 = unconvo_layer(unconv4, 512, ksize, (1, 1),'vae33',train)
    unconv6 = unconvo_layer(unconv5, 512, ksize, (1, 1),'vae34',train)
    unconv7 = unconvo_layer(unconv6, 512, ksize, (1, 1),'vae35',train)
    unconv8 = unconvo_layer(unconv7, 512, 2, (2, 2),'vae36',train)
    unconv9 = unconvo_layer(unconv8, 256, ksize, (1, 1),'vae37',train)
    unconv10 = unconvo_layer(unconv9, 256, ksize, (1, 1),'vae38',train)
    unconv11 = unconvo_layer(unconv10, 256, ksize, (1, 1),'vae39',train)
    unconv12 = unconvo_layer(unconv11, 256, 2, (2, 2),'vae40',train)
    unconv13 = unconvo_layer(unconv12, 128, ksize, (1, 1),'vae41',train)
    unconv14 = unconvo_layer(unconv13, 128, ksize, (1, 1),'vae42',train)
    unconv15 = unconvo_layer(unconv14, 128, ksize, (1, 1),'vae43',train)
    unconv16 = unconvo_layer(unconv15, 128, 2, (2, 2),'vae44',train)
    unconv17 = unconvo_layer(unconv16, 64, ksize, (1, 1),'vae45',train)
    unconv18 = unconvo_layer(unconv17, 64, ksize, (1, 1),'vae46',train)
    output = tf.layers.conv2d_transpose(unconv18, 3, 3, strides=(1, 1), padding="SAME",name='vae47',trainable=train)

    return output


def print_info(mean, std, epoch, beta_train, rl, kdl, l, f):
    info = '-> EPOCH : ' + '{:3.0f}'.format(epoch) + \
           ' , BETA VALUE : ' + '{:4.3f}'.format(beta_train) + \
           ' , RECONSTRUCTION LOSS : ' + '{:4.3f}'.format(rl) + \
           ' , REG LOSS : ' + '{:4.3f}'.format(kdl) + \
           ' , TOTAL LOSS : ' + '{:4.3f}'.format(l) + \
           ' , MEAN : ' + '{:4.3f}'.format(np.mean(mean)) + \
           ' , STD : ' + '{:4.3f}'.format(np.mean(np.exp(std / 2.)))
    print(info)
    f.write(info + '\n')