from __future__ import division, print_function, absolute_import
from tensorflow.examples.tutorials.mnist import input_data
from tensorflow.python.util import deprecation
import matplotlib.pyplot as plt
import tensorflow as tf
import numpy as np
import os

# Define VAE Loss
def vae_loss(x_reconstructed, x_true, beta, z_mean, z_std):
    # Reconstruction loss
    x_reconstructed = tf.layers.flatten(x_reconstructed)
    x_true = tf.layers.flatten(x_true)
    l2_loss = tf.reduce_sum(tf.pow(x_true-x_reconstructed,2),1)
    latent_loss = -0.5 * tf.reduce_sum(1 + z_std
                                       - tf.square(z_mean)
                                       - tf.exp(z_std), 1)
    return tf.reduce_mean(l2_loss + 2 * latent_loss ), tf.reduce_mean(l2_loss), tf.reduce_mean(latent_loss)


def file_to_print(dirName):
    try:
        # Create target Directory
        os.mkdir(dirName)
        os.mkdir(dirName + '/recons')
        os.mkdir(dirName + '/plots')
        os.mkdir(dirName + '/save')
        print("Directory ", dirName, " Created ")
        f = open(dirName + "/log.txt", "w+")
        # info = 'con batch norm en to do, si restnet , max cada 3 , 32-512(3), relu , 1024 en fc '
        # f.write(info + '\n')
    except FileExistsError:
        print("Directory ", dirName, " already exists")
    return f



def convo_layer_max(input, nfilters, ksize,strides):
    layer_norm = tf.layers.conv2d(input, nfilters, ksize, padding="SAME")
    max_layer = tf.nn.max_pool(layer_norm, ksize=[1, ksize, ksize, 1], strides=[1, strides[0], strides[1], 1], padding='SAME')
    return max_layer

def fc_layer(input, num_outputs, rate):
    layer_norm = tf.layers.dense(input, num_outputs)
    drop = tf.nn.dropout(layer_norm, rate)
    activation_layer = tf.nn.elu(drop)
    return activation_layer


def fc_layer_nodrop(input, num_outputs):
    layer_norm = tf.layers.dense(input, num_outputs)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer

def convo_layer(input, nfilters, ksize,strides):
    layer_norm = tf.layers.conv2d(input, nfilters,ksize,strides=strides, padding="SAME")
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer

def unconvo_layer(input, nfilters, ksize,  strides ):
    layer_norm = tf.layers.conv2d_transpose(input, nfilters, ksize, strides=strides, padding="SAME")
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def encoder(input,d_rate):
    ksize=3
    conv1 = convo_layer(input, 16, ksize, (1,1))
    conv2 = convo_layer_max(conv1, 16, 2, (2,2))
    conv3 = convo_layer(conv2, 32, ksize, (1,1))
    conv4 = convo_layer_max(conv3, 32, 2, (2,2))
    conv5 = convo_layer(conv4, 64, ksize, (1,1))
    conv6 = convo_layer_max(conv5, 64, 2, (2,2))
    conv7 = convo_layer(conv6, 128, ksize, (1,1))
    conv8 = convo_layer_max(conv7, 128,2, (2,2))
    conv9 = convo_layer(conv8, 256, ksize, (1,1))
    conv10 = convo_layer_max(conv9, 256, 2, (2,2))
    conv11 = convo_layer(conv10, 512, ksize, (1, 1))
    conv12 = convo_layer(conv11, 512, ksize, (1,1))
    conv13 = convo_layer(conv12, 1024, ksize, (1,1))
    conv14 = convo_layer(conv13, 1024, ksize, (1,1))

    avg = tf.compat.v1.layers.flatten(conv14)
    fc1 = fc_layer(avg, 4096, d_rate)
    fc2 = fc_layer(fc1, 4096, d_rate)

    fc3 = fc_layer_nodrop(fc2, 1024)

    return fc3

def latent_space(input, latent_dim):
    mean_ = tf.layers.dense(input, units=latent_dim, name='mean')  # , activation=tf.nn.sigmoid
    std_dev = tf.nn.softplus(tf.layers.dense(input, units=latent_dim), name='std_dev') + 1e-6
    return mean_, std_dev

def decoder(input, d_rate):
    ksize = 3
    fc0 = fc_layer_nodrop(input, 1024)
    fc1 = fc_layer(fc0, 4096, d_rate)
    fc2 = fc_layer(fc1, 4096, d_rate)

    fc3 = fc_layer_nodrop(fc2, 4 * 4 * 1024)
    rsl = tf.reshape(fc3, shape=[-1, 4, 4, 1024])

    conv1 = unconvo_layer(rsl, 1024, ksize, (1, 1))
    conv2 = unconvo_layer(conv1, 1024, ksize, (1, 1))
    conv3 = unconvo_layer(conv2, 512, ksize, (1, 1))
    conv4 = unconvo_layer(conv3, 512, ksize, (1, 1))
    conv5 = unconvo_layer(conv4, 256, ksize, (1, 1))
    conv6 = unconvo_layer(conv5, 256, ksize, (2, 2))
    conv7 = unconvo_layer(conv6, 128, ksize, (1, 1))
    conv8 = unconvo_layer(conv7, 128, ksize, (2, 2))
    conv9 = unconvo_layer(conv8, 64, ksize, (1, 1))
    conv10 = unconvo_layer(conv9, 64, ksize, (2, 2))
    conv11 = unconvo_layer(conv10, 32, ksize, (1, 1))
    conv12 = unconvo_layer(conv11, 32, ksize, (2, 2))
    conv13 = unconvo_layer(conv12, 16, ksize, (1, 1))
    conv14 = unconvo_layer(conv13, 16, ksize, (2, 2))
    unconvout = tf.layers.conv2d_transpose( conv14, 3, 3, strides=(1,1), padding="SAME")

    output = unconvout


    return output

def print_info(mean, std, epoch, beta_train, rl, kdl, l, f):
    info = '-> EPOCH : ' + '{:3.0f}'.format(epoch) + \
           ' , BETA VALUE : ' + '{:4.3f}'.format(beta_train) + \
           ' , RECONSTRUCTION LOSS : ' + '{:4.3f}'.format(rl) + \
           ' , KL LOSS : ' + '{:4.3f}'.format(kdl) + \
           ' , TOTAL LOSS : ' + '{:4.3f}'.format(l) + \
           ' , MEAN : ' + '{:4.3f}'.format(np.mean(mean)) + \
           ' , STD : ' + '{:4.3f}'.format(np.mean(np.exp(std / 2.)))
    print(info)
    f.write(info + '\n')
