from __future__ import division, print_function, absolute_import
from tensorflow.examples.tutorials.mnist import input_data
from tensorflow.python.util import deprecation
import matplotlib.pyplot as plt
import tensorflow as tf
import numpy as np
import os

# Define VAE Loss
def vae_loss(x_reconstructed, x_true, beta, z_mean, z_std):
    # Reconstruction loss
    x_reconstructed = tf.layers.flatten(x_reconstructed)
    x_true = tf.layers.flatten(x_true)
    l2_loss = tf.reduce_sum(tf.pow(x_true-x_reconstructed,2),1)
    latent_loss = -0.5 * tf.reduce_sum(1 + z_std
                                       - tf.square(z_mean)
                                       - tf.exp(z_std), 1)
    return tf.reduce_mean(l2_loss + beta * latent_loss ), tf.reduce_mean(l2_loss), tf.reduce_mean(latent_loss)


def file_to_print(dirName):
    try:
        # Create target Directory
        os.mkdir(dirName)
        os.mkdir(dirName + '/recons')
        os.mkdir(dirName + '/plots')
        os.mkdir(dirName + '/save')
        print("Directory ", dirName, " Created ")
        f = open(dirName + "/log.txt", "w+")
        # info = 'con batch norm en to do, si restnet , max cada 3 , 32-512(3), relu , 1024 en fc '
        # f.write(info + '\n')
    except FileExistsError:
        print("Directory ", dirName, " already exists")
    return f


def convo_layer_max(input, nfilters, ksize, strides,name):
    layer_norm = tf.layers.conv2d(input, nfilters, ksize, padding="SAME",name=name)
    max_layer = tf.nn.max_pool(layer_norm, ksize=[1, ksize, ksize, 1], strides=[1, strides[0], strides[1], 1],
                               padding='SAME')
    return max_layer


def fc_layer_nodrop(input, num_outputs,name):
    layer_norm = tf.layers.dense(input, num_outputs,name=name)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def fc_layer(input, num_outputs, rate,name):
    layer_norm = tf.layers.dense(input, num_outputs,name=name)
    drop = tf.nn.dropout(layer_norm, rate)
    activation_layer = tf.nn.elu(drop)
    return activation_layer


def convo_layer(input, nfilters, ksize, strides,name):
    layer_norm = tf.layers.conv2d(input, nfilters, ksize, strides=strides, padding="SAME",name=name)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def unconvo_layer(input, nfilters, ksize, strides,name):
    layer_norm = tf.layers.conv2d_transpose(input, nfilters, ksize, strides=strides, padding="SAME",name=name)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer

def encoder(input, d_rate):
    # input = ( input - img_mean ) / img_std
    ksize= 3

    conv0 = convo_layer(input, 64, 7, (2,2),'vae0')
    layer0 = tf.nn.elu(conv0)

    conv1 = convo_layer(layer0, 64, ksize, (1,1),'vae1')
    layer1 = tf.nn.elu(conv1)
    conv2 = convo_layer(layer1, 64, ksize,  (1,1),'vae2')
    layer2 = tf.nn.elu(conv2+layer0)
    conv3 = convo_layer(layer2, 64, ksize, (1,1),'vae3')
    layer3 = tf.nn.elu(conv3)
    conv4 = convo_layer(layer3, 64, ksize, (1, 1),'vae4')
    layer4 = tf.nn.elu(conv4+layer2)

    conv5 = convo_layer(layer4,128, ksize, (2,2),'vae5')
    layer5 = tf.nn.elu(conv5)
    conv6 = convo_layer(layer5, 128, ksize, (1, 1),'vae6')
    layer6 = tf.nn.elu(conv6 + conv5)
    conv7 = convo_layer(layer6, 128, ksize, (1, 1),'vae7')
    layer7 = tf.nn.elu(conv7)
    conv8 = convo_layer(layer7, 128, ksize, (1, 1),'vae8')
    layer8 = tf.nn.elu(conv8 + layer6)

    conv9 = convo_layer(layer8, 256, ksize, (2, 2),'vae9')
    layer9 = tf.nn.elu(conv9)
    conv10 = convo_layer(layer9, 256, ksize, (1, 1),'vae10')
    layer10 = tf.nn.elu(conv10 + conv9)
    conv11 = convo_layer(layer10, 256, ksize, (1, 1),'vae11')
    layer11 = tf.nn.elu(conv11)
    conv12 = convo_layer(layer11, 256, ksize, (1, 1),'vae12')
    layer12 = tf.nn.elu(conv12 + layer10)

    conv13 = convo_layer(layer12, 512, ksize, (2, 2),'vae13')
    layer13= tf.nn.elu(conv13)
    conv14 = convo_layer(layer13, 512, ksize, (1, 1),'vae14')
    layer14 = tf.nn.elu(conv14 + conv13)
    conv15 = convo_layer(layer14, 512, ksize, (1, 1),'vae15')
    layer15 = tf.nn.elu(conv15)
    conv16 = convo_layer(layer15, 512, ksize, (1, 1),'vae16')
    layer16 = tf.nn.elu(conv16 + layer15)

    avg = tf.compat.v1.layers.flatten(layer16)
    fc1 = fc_layer(avg, 4096, d_rate,'vae17')
    fc2 = fc_layer(fc1, 4096, d_rate,'vae18')

    fc3 = fc_layer_nodrop(fc2, 1024,'vae19')

    return fc3

def latent_space(input, latent_dim):
    mean_ = tf.layers.dense(input, units=latent_dim,name = 'vae20')  # , activation=tf.nn.sigmoid
    std_dev = tf.nn.softplus(tf.layers.dense(input, units=latent_dim,name ='vae21')) + 1e-6
    return mean_, std_dev

def decoder(input, d_rate):
    ksize = 3

    fc0 = fc_layer_nodrop(input, 1024,'vae22')
    fc1 = fc_layer(fc0, 4096, d_rate,'vae23')
    fc2 = fc_layer(fc1, 4096, d_rate,'vae24')

    fc4 = fc_layer_nodrop(fc2, 8*8*512,'vae42')
    rsl = tf.reshape(fc4, shape=[-1, 8,8,512])

    conv1 = unconvo_layer(rsl, 512, ksize, (1, 1),'vae25')
    layer1 = tf.nn.elu(conv1)
    conv2 = unconvo_layer(layer1, 512, ksize, (1, 1),'vae26')
    layer2 = tf.nn.elu(conv2 + rsl)
    conv3 = unconvo_layer(layer2, 512, ksize, (1, 1),'vae27')
    layer3 = tf.nn.elu(conv3)
    conv4 = unconvo_layer(layer3, 512, ksize, (1, 1),'vae28')
    layer4 = tf.nn.elu(conv4 + layer2)

    conv5 = unconvo_layer(layer4, 256, ksize, (2, 2),'vae29')
    layer5 = tf.nn.elu(conv5)
    conv6 = unconvo_layer(layer5, 256, ksize, (1, 1),'vae30')
    layer6 = tf.nn.elu(conv6 + conv5)
    conv7 = unconvo_layer(layer6, 256, ksize, (1, 1),'vae31')
    layer7 = tf.nn.elu(conv7)
    conv8 = unconvo_layer(layer7, 256, ksize, (1, 1),'vae32')
    layer8 = tf.nn.elu(conv8 + layer6)

    conv9 = unconvo_layer(layer8, 128, ksize, (2, 2),'vae33')
    layer9 = tf.nn.elu(conv9)
    conv10 = unconvo_layer(layer9, 128, ksize, (1, 1),'vae34')
    layer10 = tf.nn.elu(conv10 + conv9)
    conv11 = unconvo_layer(layer10, 128, ksize, (1, 1),'vae35')
    layer11 = tf.nn.elu(conv11)
    conv12 = unconvo_layer(layer11, 128, ksize, (1, 1),'vae36')
    layer12 = tf.nn.elu(conv12 + layer10)

    conv13 = unconvo_layer(layer12, 64, ksize, (2, 2),'vae37')
    layer13 = tf.nn.elu(conv13)
    conv14 = unconvo_layer(layer13, 64, ksize, (1, 1),'vae38')
    layer14 = tf.nn.elu(conv14 + conv13)
    conv15 = unconvo_layer(layer14, 64, ksize, (1, 1),'vae39')
    layer15 = tf.nn.elu(conv15)
    conv16 = unconvo_layer(layer15, 64, ksize, (1, 1),'vae40')
    layer16 = tf.nn.elu(conv16 + layer15)

    unconvout = tf.layers.conv2d_transpose(layer16, 3, 7, strides=(2,2), padding="SAME",name='vae41')

    output =  unconvout

    return output

def print_info(mean, std, epoch, beta_train, rl, kdl, l, f):
    info = '-> EPOCH : ' + '{:3.0f}'.format(epoch) + \
           ' , BETA VALUE : ' + '{:4.3f}'.format(beta_train) + \
           ' , RECONSTRUCTION LOSS : ' + '{:4.3f}'.format(rl) + \
           ' , KL LOSS : ' + '{:4.3f}'.format(kdl) + \
           ' , TOTAL LOSS : ' + '{:4.3f}'.format(l) + \
           ' , MEAN : ' + '{:4.3f}'.format(np.mean(mean)) + \
           ' , STD : ' + '{:4.3f}'.format(np.mean(np.exp(std / 2.)))
    print(info)
    f.write(info + '\n')
