from __future__ import division, print_function, absolute_import
from tensorflow.examples.tutorials.mnist import input_data
from tensorflow.python.util import deprecation
import matplotlib.pyplot as plt
import tensorflow as tf
import numpy as np
import os

ksize = 3

# Define VAE Loss
def vae_loss(x_reconstructed, x_true, beta, z_mean, z_std):
    # Reconstruction loss
    x_reconstructed = tf.layers.flatten(x_reconstructed)
    x_true = tf.layers.flatten(x_true)
    l2_loss = tf.reduce_sum(tf.pow(x_true - x_reconstructed, 2), 1)
    latent_loss = -0.5 * tf.reduce_sum(1 + z_std
                                       - tf.square(z_mean)
                                       - tf.exp(z_std), 1)
    return tf.reduce_mean(l2_loss + 2 * latent_loss), tf.reduce_mean(l2_loss), tf.reduce_mean(latent_loss)


def file_to_print(dirName):
    try:
        # Create target Directory
        os.mkdir(dirName)
        os.mkdir(dirName + '/recons')
        os.mkdir(dirName + '/plots')
        os.mkdir(dirName + '/save')
        print("Directory ", dirName, " Created ")
        f = open(dirName + "/log.txt", "w+")
        # info = 'con batch norm en to do, si restnet , max cada 3 , 32-512(3), relu , 1024 en fc '
        # f.write(info + '\n')
    except FileExistsError:
        print("Directory ", dirName, " already exists")
    return f


def convo_layer_max(input, nfilters, ksize, strides,name):
    layer_norm = tf.layers.conv2d(input, nfilters, ksize, padding="SAME",name=name)
    max_layer = tf.nn.max_pool(layer_norm, ksize=[1, ksize, ksize, 1], strides=[1, strides[0], strides[1], 1],
                               padding='SAME')
    return max_layer


def fc_layer_nodrop(input, num_outputs,name):
    layer_norm = tf.layers.dense(input, num_outputs,name=name)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def fc_layer(input, num_outputs, rate,name):
    layer_norm = tf.layers.dense(input, num_outputs,name=name)
    drop = tf.nn.dropout(layer_norm, rate)
    activation_layer = tf.nn.elu(drop)
    return activation_layer


def convo_layer(input, nfilters, ksize, strides,name):
    layer_norm = tf.layers.conv2d(input, nfilters, ksize, strides=strides, padding="SAME",name=name)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def unconvo_layer(input, nfilters, ksize, strides,name):
    layer_norm = tf.layers.conv2d_transpose(input, nfilters, ksize, strides=strides, padding="SAME",name=name)
    activation_layer = tf.nn.elu(layer_norm)
    return activation_layer


def encoder(input, d_rate):
    conv0 = convo_layer(input, 64, ksize, (1, 1),'vae0')
    conv1 = convo_layer(conv0, 64, ksize, (1, 1),'vae1')
    conv2 = convo_layer_max(conv1, 128, 2, (2, 2),'vae2')
    conv3 = convo_layer(conv2, 128, ksize, (1, 1),'vae3')
    conv4 = convo_layer(conv3, 128, ksize, (1, 1),'vae4')
    conv5 = convo_layer_max(conv4, 256, 2, (2, 2),'vae5')
    conv6 = convo_layer(conv5, 256, ksize, (1, 1),'vae6')
    conv7 = convo_layer(conv6, 256, ksize, (1, 1),'vae7')
    conv8 = convo_layer(conv7, 256, ksize, (1, 1),'vae8')
    conv9 = convo_layer_max(conv8, 512, 2, (2, 2),'vae9')
    conv10 = convo_layer(conv9, 512, ksize, (1, 1),'vae10')
    conv11 = convo_layer(conv10, 512, ksize, (1, 1),'vae11')
    conv12 = convo_layer(conv11, 512, ksize, (1, 1),'vae12')
    conv13 = convo_layer_max(conv12, 512, 2, (2, 2),'vae13')
    conv14 = convo_layer(conv13, 512, ksize, (1, 1),'vae14')
    conv15 = convo_layer(conv14, 512, ksize, (1, 1),'vae15')
    conv16 = convo_layer(conv15, 512, ksize, (1, 1),'vae16')
    conv17 = convo_layer_max(conv16, 512, 2, (2, 2),'vae17')

    avg = tf.compat.v1.layers.flatten(conv17,'vae18')
    fc0 = fc_layer(avg, 4096, d_rate,'vae19')
    fc1 = fc_layer(fc0, 4096, d_rate,'vae20')

    fc2 = fc_layer_nodrop(fc1, 1024,'vae21')

    return fc2


def latent_space(input, latent_dim):
    mean_ = tf.layers.dense(input, units=latent_dim, name='vae22')  # , activation=tf.nn.sigmoid
    std_dev = tf.nn.softplus(tf.layers.dense(input, units=latent_dim,name='vae23'), name='std_dev') + 1e-6
    return mean_, std_dev


def decoder(input, d_rate):
    fc0 = fc_layer_nodrop(input, 1024,'vae24')
    fc1 = fc_layer(fc0, 4096, d_rate,'vae25')
    fc2 = fc_layer(fc1, 4096, d_rate,'vae26')

    fc3 = fc_layer_nodrop(fc2, 4 * 4 * 512,'vae27')
    rsl = tf.reshape(fc3, shape=[-1, 4, 4, 512])

    unconv0 = unconvo_layer(rsl, 512, 2, (2, 2),'vae28')
    unconv1 = unconvo_layer(unconv0, 512, ksize, (1, 1),'vae29')
    unconv2 = unconvo_layer(unconv1, 512, ksize, (1, 1),'vae30')
    unconv3 = unconvo_layer(unconv2, 512, ksize, (1, 1),'vae31')
    unconv4 = unconvo_layer(unconv3, 512, 2, (2, 2),'vae32')
    unconv5 = unconvo_layer(unconv4, 512, ksize, (1, 1),'vae33')
    unconv6 = unconvo_layer(unconv5, 512, ksize, (1, 1),'vae34')
    unconv7 = unconvo_layer(unconv6, 512, ksize, (1, 1),'vae35')
    unconv8 = unconvo_layer(unconv7, 512, 2, (2, 2),'vae36')
    unconv9 = unconvo_layer(unconv8, 256, ksize, (1, 1),'vae37')
    unconv10 = unconvo_layer(unconv9, 256, ksize, (1, 1),'vae38')
    unconv11 = unconvo_layer(unconv10, 256, ksize, (1, 1),'vae39')
    unconv12 = unconvo_layer(unconv11, 256, 2, (2, 2),'vae40')
    unconv13 = unconvo_layer(unconv12, 128, ksize, (1, 1),'vae41')
    unconv14 = unconvo_layer(unconv13, 128, ksize, (1, 1),'vae42')
    unconv15 = unconvo_layer(unconv14, 128, ksize, (1, 1),'vae43')
    unconv16 = unconvo_layer(unconv15, 128, 2, (2, 2),'vae44')
    unconv17 = unconvo_layer(unconv16, 64, ksize, (1, 1),'vae45')
    unconv18 = unconvo_layer(unconv17, 64, ksize, (1, 1),'vae46')
    output = tf.layers.conv2d_transpose(unconv18, 3, 3, strides=(1, 1), padding="SAME",name='vae47')
    return output


def print_info(mean, std, epoch, beta_train, rl, kdl, l, f):
    info = '-> EPOCH : ' + '{:3.0f}'.format(epoch) + \
           ' , BETA VALUE : ' + '{:4.3f}'.format(beta_train) + \
           ' , RECONSTRUCTION LOSS : ' + '{:4.3f}'.format(rl) + \
           ' , KL LOSS : ' + '{:4.3f}'.format(kdl) + \
           ' , TOTAL LOSS : ' + '{:4.3f}'.format(l) + \
           ' , MEAN : ' + '{:4.3f}'.format(np.mean(mean)) + \
           ' , STD : ' + '{:4.3f}'.format(np.mean(np.exp(std / 2.)))
    print(info)
    f.write(info + '\n')
