from __future__ import division, print_function, absolute_import
import matplotlib
matplotlib.use('Agg')
from tensorflow.python.util import deprecation
import matplotlib.pyplot as plt
import tensorflow as tf
import numpy as np
import os
import cv2 as cv2
import glob as glob
from tools_2part_resnet import encoder, decoder, print_info, file_to_print, latent_space, vae_loss
import random
from numpy import load
from sklearn.manifold import TSNE
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.contrib._warning = None
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = "4"

image_dim = 128
# Training parameters
learning_rate = 1e-5
batch_size = 16
epochs = 1500
dropout_rate = 0.5
latent_dim = 64
num_images = 7
path_to_decoder = '/src/datasets/bestMse2/bestmodel.ckpt'
# path_to_decoder = '/src/datasets/bestMse/bestmodel.ckpt'
# Testing Parameters
samples = 30 # number of faces to be printed
reconstruction_images = 8
min_val = 1000000
min_epoch=0
# Saving and loading parameters
num_training_samples=544
location_to_save = '/src/models'
network_name = 'REAL_SIMILAR_RESNET_REG'
directory_name = location_to_save + '/' + network_name
path_to_dataset = '/src/datasets/FINAL2'
selected_subjects =['2q8orkMs2Jg.000','9UEBU_yiTPs.001',
                    'EHPZNxP5mN8.000','be0DQawtVkE.000',
                    'Cng72emlXAM.004','BLeDNfK2quI.003',
                    'c2LeNguVNrI.003','J6RHl6U6Jcw.000']

# Charge data
entries = os.listdir(path_to_dataset)
neutral_faces = []
frontal_faces = []
neutral_paths = []
regularization_vectors = []
entries = random.sample(entries, len(entries))

for i in range(reconstruction_images):
    index = entries.index(selected_subjects[i])
    aux = entries[num_training_samples+i]
    entries[index] = aux
    entries[num_training_samples+i] = selected_subjects[i]


for directory in entries:
    vect = glob.glob(path_to_dataset + '/' + directory + '/data_incr2.npy')
    # vect = glob.glob(path_to_dataset + '/' + directory + '/data_beta.npy')
    vector = load(vect[0])
    regularization_vectors.append(vector)



for directory in entries:
    filenames = [img for img in glob.glob(path_to_dataset + '/' + directory + '/*.png')]
    count = 0
    neutral_found = False
    neutral = ''
    frontals = []
    for file in filenames:
        if os.path.basename(file)[0] == 'N':
            neutral = file
            neutral_found = True
        if os.path.basename(file)[0] == 'F' and count < num_images:
            frontals.append(cv2.imread(file) / 255.)
            count = count + 1
    if neutral_found and count >= num_images :
        neutral_faces.append(cv2.imread(neutral) / 255.)
        neutral_paths.append(neutral)
        frontal_faces.append( frontals )

print('---> Total length dataset : ',len(neutral_faces))
training_set_neutral = neutral_faces[0:num_training_samples]
training_set_frontal_faces = frontal_faces[0:num_training_samples]
training_set_regularization_vectors = regularization_vectors[0:num_training_samples]
print('---> Total training size: ',len(training_set_neutral))
test_set_neutral = neutral_faces[num_training_samples:]
test_set_frontal_faces = frontal_faces[num_training_samples:]
test_set_regularization_vectors = regularization_vectors[num_training_samples:]
print('---> Total test size: ',len(test_set_neutral))
print('---> Batch size : ',batch_size)
total_batchs = int(len(training_set_neutral) / batch_size)
total_batchs_test = int(len(test_set_neutral) / batch_size)
print("---> Total Train Batch Size : ", total_batchs)
print("---> Total Test Size : ", len(test_set_neutral))
neutral_paths_test = neutral_paths[num_training_samples:]
print('Total regularization vectors : ',len(regularization_vectors))

#########################################   MODEL   ##########################################
# Placeholders
X_true = tf.placeholder("float", [None, image_dim, image_dim, 3])
X_faces = tf.placeholder("float", [None, num_images, image_dim, image_dim, 3])
Z_sampled = tf.placeholder("float", [None, latent_dim])
Reg_vector = tf.placeholder("float", [None, latent_dim])
beta = tf.placeholder(tf.float32, [], name='beta')
d_rate = tf.placeholder_with_default(1.0, shape=(), name='dropout')
# Encode
enco = encoder(X_faces, d_rate,num_images)
# Generate distribution
z_mean, z_std = latent_space(enco, latent_dim)
eps = tf.random_normal(tf.shape(z_std), dtype=tf.float32, mean=0., stddev=1.0, name='epsilon')
# Sampler: Normal (gaussian) random distribution
# Reparametrization trick
Z_sampled = z_mean + tf.exp(z_std / 2.) * eps
# Decode
deco = decoder(Z_sampled, d_rate)
# Loss
loss, reconstruction_loss, kdloss = vae_loss(deco, X_true, beta, z_mean, z_std,Z_sampled,Reg_vector)
# Optimizer
train_op = tf.train.AdamOptimizer(learning_rate).minimize(loss)
# Initialize variables
init = tf.global_variables_initializer()
##########################################   FINISH MODEL   #######################################

def getImage(path):
    return OffsetImage(plt.imread(path), zoom=0.25)

# Start Training
# Start a new TF session
with tf.Session() as sess:
    # Run the initializer
    sess.run(init)
    # Restore decoder
    # reuse_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES,
    #                                scope="vae[0123456789]")
    # reuse_vars_dict = dict([(var.op.name, var) for var in reuse_vars])
    # restore_saver = tf.train.Saver(reuse_vars_dict)
    # restore_saver.restore(sess, path_to_decoder)
    # TRAINING
    stop = False
    epoch = 1
    beta_train = 0.0
    f = file_to_print(directory_name)
    # Some images that will be reconstructed, to visually check quality
    for i in range(reconstruction_images):
        cv2.imwrite(directory_name + '/recons/original_img' + i.__str__() + '_epoch_' + epoch.__str__() + '.png',
                    (test_set_neutral[i]) * 255.)
    # Training variables for info
    total_rl = 0.
    total_kdl = 0.
    total_l = 0.
    while epoch < epochs + 1 and (not stop):
        # More training variables for info
        mean = np.zeros(latent_dim, float)
        std = np.zeros(latent_dim, float)
        total_rl = 0.
        total_kdl = 0.
        total_l = 0.
        # Batch training
        for i in range(total_batchs):
            reg_vect_batch = training_set_regularization_vectors[i * batch_size:((i + 1) * batch_size)]
            frontal_batch = training_set_frontal_faces[i * batch_size:((i + 1) * batch_size)]
            neutral_batch = training_set_neutral[i * batch_size:((i + 1) * batch_size)]
            _, l, rl, kdl, m, s = sess.run([train_op, loss, reconstruction_loss, kdloss, z_mean, z_std],
                                           feed_dict={X_faces: frontal_batch,
                                                      X_true: neutral_batch,
                                                      Reg_vector: reg_vect_batch,
                                                      beta: beta_train,
                                                      d_rate: dropout_rate})

            mean_batch = np.sum(m, axis=0) / batch_size
            std_batch = np.sum(s, axis=0) / batch_size
            mean += mean_batch
            std += std_batch
            total_rl += (rl / batch_size)
            total_kdl += (kdl / batch_size)
            total_l += (l / batch_size)

        total_rl /= total_batchs
        total_kdl /= total_batchs
        total_l /= total_batchs
        mean /= total_batchs
        std /= total_batchs

        print_info(mean, std, epoch, beta_train, total_rl, total_kdl, total_l, f)

        # Plotting distribution of the validation set with T-SNE
        validation_list = np.empty([latent_dim, latent_dim])
        total_rlv = 0
        if (epoch % 5 == 0):
            for i in range(total_batchs_test):
                frontal_batch = test_set_frontal_faces[i * batch_size:((i + 1) * batch_size)]
                neutral_batch = test_set_neutral[i * batch_size:((i + 1) * batch_size)]
                reconstruction, rlvalidation, z = sess.run([deco, reconstruction_loss, Z_sampled],
                                                           feed_dict={X_faces: frontal_batch,
                                                                      X_true: neutral_batch,
                                                                      beta: beta_train})
                total_rlv += (rlvalidation / batch_size)
                validation_list = np.concatenate((validation_list, z))

            total_rlv /= total_batchs_test

            reduced = TSNE(n_components=2).fit_transform(validation_list[latent_dim:])

            x = [p[0] for p in reduced]
            y = [p[1] for p in reduced]

            for z in range(10):
                fig, ax = plt.subplots()
                plt.plot(x, y, 'bo')
                plt.savefig(directory_name + '/plots/plot_' + z.__str__() + '_points_epoch_' + epoch.__str__() + '.png')
                random_numbers = random.sample(range(total_batchs_test * batch_size), 64)

                for k in random_numbers:
                    ab = AnnotationBbox(getImage(neutral_paths_test[k]), (x[k], y[k]), frameon=False)
                    ax.add_artist(ab)
                plt.savefig(directory_name + '/plots/plot_' + z.__str__() + '_images_epoch_' + epoch.__str__() + '.png')

            # Reconstruction Images
            reconstruction = sess.run(deco,
                                      feed_dict={X_faces: test_set_frontal_faces[0:reconstruction_images],
                                                 beta: beta_train})
            for i in range(reconstruction_images):
                cv2.imwrite(
                    directory_name + '/recons/' + epoch.__str__() + '_epoch_' + 'reconstructed_img' + i.__str__() + '.png',
                    reconstruction[i] * 255.)

            validation_loss = total_rlv
            print('-> MSE : ')
            print(validation_loss)
            f.write('VALIDATION LOSS : ' + str(validation_loss) + '\n')
            diff = np.abs(validation_loss - total_rl) / validation_loss
            print('DIFF ERROR : ' + str(diff) + '\n')
            f.write('DIFF ERROR : ' + str(diff) + '\n')

            if epoch > 1 and diff >= 0.30:
                stop = True
            if ( validation_loss < min_val ) :
                min_val = validation_loss
                min_epoch = epoch

        epoch += 1

    f.write('Min MSE : ' + min_val.__str__() + '\n' + 'Epoch : ' + min_epoch.__str__())
    print('---> FINISHED_TRAINING')
