from __future__ import division, print_function, absolute_import
import matplotlib
matplotlib.use('Agg')
from tensorflow.python.util import deprecation
import matplotlib.pyplot as plt
import tensorflow as tf
import numpy as np
import os
import cv2 as cv2
import glob as glob
from tools_resnet import encoder, decoder, print_info, file_to_print, latent_space, vae_loss
import random
from sklearn.manifold import TSNE
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.contrib._warning = None
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = "6"

image_dim = 128
# Training parameters
numTraining = 4928
learning_rate = 1e-5
batch_size = 64
epochs = 1500
dropout_rate = 0.75
latent_dim = 64
# Testing Parameters
samples = 30 # number of faces to be printed
reconstruction_images = 8
faces_to_plot = 55
selected_images=[29,52,151,2595,194,2640,761,3052]
min_val = 1000000
min_epoch=0
# Saving and loading parameters
location_to_save = '/src/models'
network_name = 'RESNET'
directory_name = location_to_save + '/' + network_name
path_to_dataset = '/src/datasets/FINAL/'

# Charge data
filenames = [img for img in glob.glob(path_to_dataset + '*.png')]
filenames.sort()
filenames = random.sample(filenames, len(filenames))
# Organizing selected images for reconstruction plotting
for i in range(reconstruction_images):
    index = filenames.index(path_to_dataset+selected_images[i].__str__()+'.png')
    aux = filenames[numTraining+i]
    filenames[index] = aux
    filenames[numTraining+i] = path_to_dataset+selected_images[i].__str__()+'.png'

test_set_names = filenames[numTraining:]
images = [cv2.imread(img) / 255. for img in filenames]
training_set = images[0:numTraining]
test_set = images[numTraining:]

##########################################   MODEL   ##########################################
# Placeholders
X_true = tf.placeholder("float", [None, image_dim, image_dim, 3])
Z_sampled = tf.placeholder("float", [None, latent_dim])
beta = tf.placeholder(tf.float32, [], name='beta')
d_rate = tf.placeholder_with_default(1.0, shape=(), name='dropout')
# Encode
enco = encoder(X_true, d_rate)
# Generate distribution
z_mean, z_std = latent_space(enco, latent_dim)
eps = tf.random_normal(tf.shape(z_std), dtype=tf.float32, mean=0., stddev=1.0, name='epsilon')
# Sampler: Normal (gaussian) random distribution
# Reparametrization trick
Z_sampled = z_mean + tf.exp(z_std / 2.) * eps
# Decode
deco = decoder(Z_sampled, d_rate)
# Loss
loss, reconstruction_loss, kdloss = vae_loss(deco, X_true, beta, z_mean, z_std)
# Optimizer
train_op = tf.train.AdamOptimizer(learning_rate).minimize(loss)
# Initialize variables
init = tf.global_variables_initializer()
##########################################   FINISH MODEL   #######################################

def getImage(path):
    return OffsetImage(plt.imread(path), zoom=0.25)

# Start Training
# Start a new TF session
with tf.Session() as sess:
    # Run the initializer
    sess.run(init)
    # Checking lenghts
    total_batchs = int(len(training_set) / batch_size)
    total_batchs_test = int(len(test_set) / batch_size)
    print("---> Total Train Batch Size : ", total_batchs)
    print("---> Total Train Size : ", len(training_set))
    print("---> Total Test Batch Size : ", total_batchs_test)
    print("---> Total Test Size : ", len(test_set))
    # TRAINING
    stop = False
    epoch = 1
    beta_train = 0.001
    f = file_to_print(directory_name)
    # Some images that will be reconstructed, to visually check quiality
    for i in range(reconstruction_images):
        cv2.imwrite(directory_name + '/recons/original_img' + i.__str__() + '_epoch_' + epoch.__str__() + '.png',
                    (test_set[i]) * 255.)
    # Training variables for info
    total_rl = 0.
    total_kdl = 0.
    total_l = 0.
    while epoch < epochs + 1 and (not stop):
        # More training variables for info
        mean = np.zeros(latent_dim, float)
        std = np.zeros(latent_dim, float)
        total_rl = 0.
        total_kdl = 0.
        total_l = 0.
        # Batch training
        for i in range(total_batchs):
            batch_x = training_set[i * batch_size:((i + 1) * batch_size)]

            _, l, rl, kdl, m, s = sess.run([train_op, loss, reconstruction_loss, kdloss, z_mean, z_std],
                                           feed_dict={X_true: batch_x, beta: beta_train, d_rate: dropout_rate})

            mean_batch = np.sum(m, axis=0) / batch_size
            std_batch = np.sum(s, axis=0) / batch_size
            mean += mean_batch
            std += std_batch
            total_rl += (rl / batch_size)
            total_kdl += (kdl / batch_size)
            total_l += (l / batch_size)

        total_rl /= total_batchs
        total_kdl /= total_batchs
        total_l /= total_batchs
        mean /= total_batchs
        std /= total_batchs

        print_info(mean, std, epoch, beta_train, total_rl, total_kdl, total_l, f)

        # Plotting distribution of the validation set with T-SNE
        validation_list = np.empty([latent_dim, latent_dim])
        total_rlv = 0
        if (epoch % 5 == 0):
            for i in range(total_batchs_test):
                batch_x = test_set[i * batch_size:((i + 1) * batch_size)]
                reconstruction, rlvalidation, z = sess.run([deco, reconstruction_loss, Z_sampled],
                                                           feed_dict={X_true: test_set[0:batch_size], beta: 0.0})
                total_rlv += (rlvalidation / batch_size)
                validation_list = np.concatenate((validation_list, z))

            total_rlv /= total_batchs_test

            reduced = TSNE(n_components=2,perplexity=45,learning_rate=300.0,n_iter=2000).fit_transform(validation_list[latent_dim:])

            x = [p[0] for p in reduced]
            y = [p[1] for p in reduced]

            for z in range(10):
                fig, ax = plt.subplots()
                plt.plot(x, y, 'bo')
                random_numbers = random.sample(range(total_batchs_test * batch_size), faces_to_plot)
                for k in random_numbers:
                    ab = AnnotationBbox(getImage(test_set_names[k]), (x[k], y[k]), frameon=False)
                    ax.add_artist(ab)
                plt.savefig(directory_name + '/plots/plot_' + z.__str__() + '_images_epoch_' + epoch.__str__() + '.png')
                plt.close('all')
            # Reconstruction Images
            reconstruction = sess.run(deco, feed_dict={X_true: test_set[0:reconstruction_images], beta: 0.0})
            for i in range(reconstruction_images):
                cv2.imwrite(
                    directory_name + '/recons/' + epoch.__str__() + '_epoch_' + 'reconstructed_img' + i.__str__() + '.png',
                    reconstruction[i] * 255.)

            validation_loss = total_rlv
            print('-> MSE : ')
            print(validation_loss)
            f.write('VALIDATION LOSS : ' + str(validation_loss) + '\n')
            diff = np.abs(validation_loss - total_rl) / validation_loss
            print('DIFF ERROR : ' + str(diff) + '\n')
            f.write('DIFF ERROR : ' + str(diff) + '\n')
            if (validation_loss < 10.0 and beta_train < 1.0):
                beta_train = beta_train + 0.05
            elif epoch > 1 and diff >= 0.30:
                stop = True

            if ( validation_loss < min_val and beta_train > 1.0) :
                min_val = validation_loss
                min_epoch = epoch
                tf.train.Saver().save(sess, directory_name + '/save/bestmodel.ckpt')


        # Sampling from noise and distribution
        if (epoch % 25 == 0):
            random_vector = np.random.uniform(low=0.0, high=1.0, size=[samples, image_dim * image_dim * 3])
            # Encode and decode the digit image
            g = sess.run(deco, feed_dict={X_true: random_vector.reshape([samples, image_dim, image_dim, 3]), beta: 0.0})
            for j in range(samples):
                cv2.imwrite(
                    directory_name + '/' + epoch.__str__() + '_epoch_' + 'generated_from_noise_img' + j.__str__() + '.png',
                    g[j] * 255.)
            print('PRINTED IMAGES SAMPLED FROM NOISE')

            std_matrix = np.zeros((latent_dim, latent_dim), float)
            np.fill_diagonal(std_matrix, 1, wrap=True)
            random_normal = np.random.multivariate_normal(np.zeros(latent_dim, float), std_matrix, samples)
            # Encode and decode the digit image
            g = sess.run(deco, feed_dict={Z_sampled: random_normal.reshape([samples, latent_dim]), beta: 0.0})
            for j in range(samples):
                cv2.imwrite(
                    directory_name + '/' + epoch.__str__() + '_epoch_' + 'generated_from_latent_img' + j.__str__() + '.png',
                    g[j] * 255.)
            print('PRINTED IMAGES SAMPLED FROM Z')

        epoch += 1
    f.write('Min MSE : ' + min_val.__str__() + '\n'+'Epoch : '+min_epoch.__str__())
    print('---> FINISHED_TRAINING')
